/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: ftdecode.c,v 1.3 2001/02/10 06:43:30 maf Exp $
 */

#include <strings.h>
#include "ftlib.h"

/*
 * function: fts2rec_pdu_decode
 *
 * pdu must be in network byte order
 * stream records are returned in the byte order defined by bo
 * XXX multiple functions
 * XXX verify PDU sanity
 * XXX ftdecode needs a jump table
 *
 * returns: # of stream records decoded
*/
int fts2rec_pdu_decode(void *pdu, struct ftdecode *decode, int bo)
{
  int n;
  struct ftpdu_header *ph;
  struct ftpdu_v1 *pdu_v1;
  struct ftpdu_v5 *pdu_v5;
  struct ftpdu_v6 *pdu_v6;
  struct ftpdu_v7 *pdu_v7;
  struct ftpdu_v8_1 *pdu_v8_1;
  struct ftpdu_v8_2 *pdu_v8_2;
  struct ftpdu_v8_3 *pdu_v8_3;
  struct ftpdu_v8_4 *pdu_v8_4;
  struct ftpdu_v8_5 *pdu_v8_5;
  struct fts2rec_v1 *rec_v1;
  struct fts2rec_v5 *rec_v5;
  struct fts2rec_v6 *rec_v6;
  struct fts2rec_v7 *rec_v7;
  struct fts2rec_v8_1 *rec_v8_1;
  struct fts2rec_v8_2 *rec_v8_2;
  struct fts2rec_v8_3 *rec_v8_3;
  struct fts2rec_v8_4 *rec_v8_4;
  struct fts2rec_v8_5 *rec_v8_5;

  n = -1;

  bzero(decode, sizeof (struct ftdecode));

  /* take advantage that all pdu's have a common header.  The
     decode sequence for all pdu headers here
  */

  ph = (struct ftpdu_header*)pdu;

#if BYTE_ORDER == LITTLE_ENDIAN
  SWAPINT16(ph->version);
  SWAPINT16(ph->count);
  SWAPINT32(ph->sysUpTime);
  SWAPINT32(ph->unix_secs);
  SWAPINT32(ph->unix_nsecs);
#endif /* LITTLE_ENDIAN */

  decode->ver.s_version = 2;
  decode->ver.d_version = ph->version;

  /*
   * ph->unix_secs, ph->unix_nsecs, ph->sysUpTime, and ph->count are in host
   * byte order.  The rest of the PDU is in network byte order
   *
   * if the target byte order is BIG_ENDIAN then unix_nsecs, unix_secs,
   * and sysUpTime, must be swapped
   *
   * if the target byte order is LITTLE_ENDIAN then the rest of the
   * PDU must be swapped
   */

  if (bo == FT_HEADER_BIG_ENDIAN) {

    SWAPINT32(ph->unix_secs);
    SWAPINT32(ph->unix_nsecs);
    SWAPINT32(ph->sysUpTime);

  }

  switch (ph->version) {

    case 1:

      decode->rec_size = ftrec_size(decode->ver);
      pdu_v1 = (struct ftpdu_v1*) pdu;

      for (n = 0; n < pdu_v1->count; ++n) {

        rec_v1 = (struct fts2rec_v1*) (decode->buf + (n*decode->rec_size));

        rec_v1->unix_nsecs = ph->unix_nsecs;
        rec_v1->unix_secs = ph->unix_secs;
        rec_v1->sysUpTime = ph->sysUpTime;

        rec_v1->srcaddr = pdu_v1->records[n].srcaddr;
        rec_v1->dstaddr = pdu_v1->records[n].dstaddr;
        rec_v1->nexthop = pdu_v1->records[n].nexthop;
        rec_v1->input = pdu_v1->records[n].input;
        rec_v1->output = pdu_v1->records[n].output;
        rec_v1->dPkts = pdu_v1->records[n].dPkts;
        rec_v1->dOctets = pdu_v1->records[n].dOctets;
        rec_v1->First = pdu_v1->records[n].First;
        rec_v1->Last = pdu_v1->records[n].Last;
        rec_v1->dstport = pdu_v1->records[n].dstport;
        rec_v1->srcport = pdu_v1->records[n].srcport;
        rec_v1->prot = pdu_v1->records[n].prot;
        rec_v1->tos = pdu_v1->records[n].tos;
        rec_v1->tcp_flags = pdu_v1->records[n].flags;

        if (bo == FT_HEADER_LITTLE_ENDIAN) {

          SWAPINT32(rec_v1->srcaddr);
          SWAPINT32(rec_v1->dstaddr);
          SWAPINT32(rec_v1->nexthop);
          SWAPINT16(rec_v1->input);
          SWAPINT16(rec_v1->output);
          SWAPINT32(rec_v1->dPkts);
          SWAPINT32(rec_v1->dOctets);
          SWAPINT32(rec_v1->First);
          SWAPINT32(rec_v1->Last);
          SWAPINT16(rec_v1->dstport);
          SWAPINT16(rec_v1->srcport);

        }

      } /* for n */

      break;

    case 5:
      decode->rec_size = ftrec_size(decode->ver);
      pdu_v5 = (struct ftpdu_v5*) pdu;

      for (n = 0; n < pdu_v5->count; ++n) {

        rec_v5 = (struct fts2rec_v5*) (decode->buf + (n*decode->rec_size));

        rec_v5->unix_nsecs = ph->unix_nsecs;
        rec_v5->unix_secs = ph->unix_secs;
        rec_v5->sysUpTime = ph->sysUpTime;

        rec_v5->engine_type = pdu_v5->engine_type;
        rec_v5->engine_type = pdu_v5->engine_id;


        rec_v5->srcaddr = pdu_v5->records[n].srcaddr;
        rec_v5->dstaddr = pdu_v5->records[n].dstaddr;
        rec_v5->nexthop = pdu_v5->records[n].nexthop;
        rec_v5->input = pdu_v5->records[n].input;
        rec_v5->output = pdu_v5->records[n].output;
        rec_v5->dPkts = pdu_v5->records[n].dPkts;
        rec_v5->dOctets = pdu_v5->records[n].dOctets;
        rec_v5->First = pdu_v5->records[n].First;
        rec_v5->Last = pdu_v5->records[n].Last;
        rec_v5->dstport = pdu_v5->records[n].dstport;
        rec_v5->srcport = pdu_v5->records[n].srcport;
        rec_v5->prot = pdu_v5->records[n].prot;
        rec_v5->tos = pdu_v5->records[n].tos;
        rec_v5->tcp_flags = pdu_v5->records[n].tcp_flags;
        rec_v5->src_as = pdu_v5->records[n].src_as;
        rec_v5->dst_as = pdu_v5->records[n].dst_as;
        rec_v5->src_mask = pdu_v5->records[n].src_mask;
        rec_v5->dst_mask = pdu_v5->records[n].dst_mask;

        if (bo == FT_HEADER_LITTLE_ENDIAN) {

          SWAPINT32(rec_v5->srcaddr);
          SWAPINT32(rec_v5->dstaddr);
          SWAPINT32(rec_v5->nexthop);
          SWAPINT16(rec_v5->input);
          SWAPINT16(rec_v5->output);
          SWAPINT32(rec_v5->dPkts);
          SWAPINT32(rec_v5->dOctets);
          SWAPINT32(rec_v5->First);
          SWAPINT32(rec_v5->Last);
          SWAPINT16(rec_v5->dstport);
          SWAPINT16(rec_v5->srcport);
          SWAPINT16(rec_v5->src_as);
          SWAPINT16(rec_v5->dst_as);

        }

      } /* for n */
      break;

    case 6:
      decode->rec_size = ftrec_size(decode->ver);
      pdu_v6 = (struct ftpdu_v6*) pdu;

      for (n = 0; n < pdu_v6->count; ++n) {

        rec_v6 = (struct fts2rec_v6*) (decode->buf + (n*decode->rec_size));

        rec_v6->unix_nsecs = ph->unix_nsecs;
        rec_v6->unix_secs = ph->unix_secs;
        rec_v6->sysUpTime = ph->sysUpTime;

        rec_v6->engine_type = pdu_v6->engine_type;
        rec_v6->engine_type = pdu_v6->engine_id;


        rec_v6->srcaddr = pdu_v6->records[n].srcaddr;
        rec_v6->dstaddr = pdu_v6->records[n].dstaddr;
        rec_v6->nexthop = pdu_v6->records[n].nexthop;
        rec_v6->input = pdu_v6->records[n].input;
        rec_v6->output = pdu_v6->records[n].output;
        rec_v6->dPkts = pdu_v6->records[n].dPkts;
        rec_v6->dOctets = pdu_v6->records[n].dOctets;
        rec_v6->First = pdu_v6->records[n].First;
        rec_v6->Last = pdu_v6->records[n].Last;
        rec_v6->dstport = pdu_v6->records[n].dstport;
        rec_v6->srcport = pdu_v6->records[n].srcport;
        rec_v6->prot = pdu_v6->records[n].prot;
        rec_v6->tos = pdu_v6->records[n].tos;
        rec_v6->tcp_flags = pdu_v6->records[n].tcp_flags;
        rec_v6->src_as = pdu_v6->records[n].src_as;
        rec_v6->dst_as = pdu_v6->records[n].dst_as;
        rec_v6->src_mask = pdu_v6->records[n].src_mask;
        rec_v6->dst_mask = pdu_v6->records[n].dst_mask;

        if (bo == FT_HEADER_LITTLE_ENDIAN) {

          SWAPINT32(rec_v6->srcaddr);
          SWAPINT32(rec_v6->dstaddr);
          SWAPINT32(rec_v6->nexthop);
          SWAPINT16(rec_v6->input);
          SWAPINT16(rec_v6->output);
          SWAPINT32(rec_v6->dPkts);
          SWAPINT32(rec_v6->dOctets);
          SWAPINT32(rec_v6->First);
          SWAPINT32(rec_v6->Last);
          SWAPINT16(rec_v6->dstport);
          SWAPINT16(rec_v6->srcport);
          SWAPINT16(rec_v6->src_as);
          SWAPINT16(rec_v6->dst_as);

        }

      } /* for n */
      break;

    case 7:
      decode->rec_size = ftrec_size(decode->ver);
      pdu_v7 = (struct ftpdu_v7*) pdu;

      for (n = 0; n < pdu_v7->count; ++n) {

        rec_v7 = (struct fts2rec_v7*) (decode->buf + (n*decode->rec_size));

        rec_v7->unix_nsecs = ph->unix_nsecs;
        rec_v7->unix_secs = ph->unix_secs;
        rec_v7->sysUpTime = ph->sysUpTime;

        rec_v7->engine_type = pdu_v7->engine_type;
        rec_v7->engine_type = pdu_v7->engine_id;

        rec_v7->srcaddr = pdu_v7->records[n].srcaddr;
        rec_v7->dstaddr = pdu_v7->records[n].dstaddr;
        rec_v7->nexthop = pdu_v7->records[n].nexthop;
        rec_v7->input = pdu_v7->records[n].input;
        rec_v7->output = pdu_v7->records[n].output;
        rec_v7->dPkts = pdu_v7->records[n].dPkts;
        rec_v7->dOctets = pdu_v7->records[n].dOctets;
        rec_v7->First = pdu_v7->records[n].First;
        rec_v7->Last = pdu_v7->records[n].Last;
        rec_v7->dstport = pdu_v7->records[n].dstport;
        rec_v7->srcport = pdu_v7->records[n].srcport;
        rec_v7->prot = pdu_v7->records[n].prot;
        rec_v7->tos = pdu_v7->records[n].tos;
        rec_v7->tcp_flags = pdu_v7->records[n].tcp_flags;
        rec_v7->src_as = pdu_v7->records[n].src_as;
        rec_v7->dst_as = pdu_v7->records[n].dst_as;
        rec_v7->src_mask = pdu_v7->records[n].src_mask;
        rec_v7->dst_mask = pdu_v7->records[n].dst_mask;
        rec_v7->router_sc = pdu_v7->records[n].router_sc;

        if (bo == FT_HEADER_LITTLE_ENDIAN) {

          SWAPINT32(rec_v7->srcaddr);
          SWAPINT32(rec_v7->dstaddr);
          SWAPINT32(rec_v7->nexthop);
          SWAPINT16(rec_v7->input);
          SWAPINT16(rec_v7->output);
          SWAPINT32(rec_v7->dPkts);
          SWAPINT32(rec_v7->dOctets);
          SWAPINT32(rec_v7->First);
          SWAPINT32(rec_v7->Last);
          SWAPINT16(rec_v7->dstport);
          SWAPINT16(rec_v7->srcport);
          SWAPINT16(rec_v7->src_as);
          SWAPINT16(rec_v7->dst_as);
          SWAPINT32(rec_v7->router_sc);

        }
      } /* for */

      break;

    case 8:
      /* All v8 exports have a common sub header.  decode it here for
         all aggregation methods
      */

      decode->ver.agg_method = *((char*)pdu+22);
      decode->ver.agg_version = *((char*)pdu+23);

      if (decode->ver.agg_version != 2) {
        fprintf(stderr, "Unsupported agg_version\n");
        goto fts2rec_pdu_decode_out;
      }

      switch (decode->ver.agg_method) {

        case 1:
          decode->rec_size = ftrec_size(decode->ver);
          pdu_v8_1 = (struct ftpdu_v8_1*) pdu;

          for (n = 0; n < pdu_v8_1->count; ++n) {

            rec_v8_1 = (struct fts2rec_v8_1*)
              (decode->buf + (n*decode->rec_size));

            rec_v8_1->unix_nsecs = ph->unix_nsecs;
            rec_v8_1->unix_secs = ph->unix_secs;
            rec_v8_1->sysUpTime = ph->sysUpTime;

            rec_v8_1->engine_type = pdu_v8_1->engine_type;
            rec_v8_1->engine_type = pdu_v8_1->engine_id;

            rec_v8_1->dFlows = pdu_v8_1->records[n].dFlows;
            rec_v8_1->dPkts = pdu_v8_1->records[n].dPkts;
            rec_v8_1->dOctets = pdu_v8_1->records[n].dOctets;
            rec_v8_1->First = pdu_v8_1->records[n].First;
            rec_v8_1->Last = pdu_v8_1->records[n].Last;
            rec_v8_1->src_as = pdu_v8_1->records[n].src_as;
            rec_v8_1->dst_as = pdu_v8_1->records[n].dst_as;
            rec_v8_1->input = pdu_v8_1->records[n].input;
            rec_v8_1->output = pdu_v8_1->records[n].output;

            if (bo == FT_HEADER_LITTLE_ENDIAN) {

              SWAPINT32(rec_v8_1->dFlows);
              SWAPINT32(rec_v8_1->dPkts);
              SWAPINT32(rec_v8_1->dOctets);
              SWAPINT32(rec_v8_1->First);
              SWAPINT32(rec_v8_1->Last);
              SWAPINT16(rec_v8_1->src_as);
              SWAPINT16(rec_v8_1->dst_as);
              SWAPINT16(rec_v8_1->input);
              SWAPINT16(rec_v8_1->output);

            }

          } /* for */

          break;
    
        case 2:
          decode->rec_size = ftrec_size(decode->ver);
          pdu_v8_2 = (struct ftpdu_v8_2*) pdu;

          for (n = 0; n < pdu_v8_2->count; ++n) {

            rec_v8_2 = (struct fts2rec_v8_2*)
              (decode->buf + (n*decode->rec_size));

            rec_v8_2->unix_nsecs = ph->unix_nsecs;
            rec_v8_2->unix_secs = ph->unix_secs;
            rec_v8_2->sysUpTime = ph->sysUpTime;

            rec_v8_2->engine_type = pdu_v8_2->engine_type;
            rec_v8_2->engine_type = pdu_v8_2->engine_id;

            rec_v8_2->dFlows = pdu_v8_2->records[n].dFlows;
            rec_v8_2->dPkts = pdu_v8_2->records[n].dPkts;
            rec_v8_2->dOctets = pdu_v8_2->records[n].dOctets;
            rec_v8_2->First = pdu_v8_2->records[n].First;
            rec_v8_2->Last = pdu_v8_2->records[n].Last;
            rec_v8_2->srcport = pdu_v8_2->records[n].srcport;
            rec_v8_2->dstport = pdu_v8_2->records[n].dstport;
            rec_v8_2->prot = pdu_v8_2->records[n].prot;

            if (bo == FT_HEADER_LITTLE_ENDIAN) {

              SWAPINT32(rec_v8_2->dFlows);
              SWAPINT32(rec_v8_2->dPkts);
              SWAPINT32(rec_v8_2->dOctets);
              SWAPINT32(rec_v8_2->First);
              SWAPINT32(rec_v8_2->Last);
              SWAPINT16(rec_v8_2->srcport);
              SWAPINT16(rec_v8_2->dstport);

            }

          } /* for */

          break;

        case 3:
          decode->rec_size = ftrec_size(decode->ver);
          pdu_v8_3 = (struct ftpdu_v8_3*) pdu;

          for (n = 0; n < pdu_v8_3->count; ++n) {

            rec_v8_3 = (struct fts2rec_v8_3*)
              (decode->buf + (n*decode->rec_size));

            rec_v8_3->unix_nsecs = ph->unix_nsecs;
            rec_v8_3->unix_secs = ph->unix_secs;
            rec_v8_3->sysUpTime = ph->sysUpTime;

            rec_v8_3->engine_type = pdu_v8_3->engine_type;
            rec_v8_3->engine_type = pdu_v8_3->engine_id;

            rec_v8_3->dFlows = pdu_v8_3->records[n].dFlows;
            rec_v8_3->dPkts = pdu_v8_3->records[n].dPkts;
            rec_v8_3->dOctets = pdu_v8_3->records[n].dOctets;
            rec_v8_3->First = pdu_v8_3->records[n].First;
            rec_v8_3->Last = pdu_v8_3->records[n].Last;
            rec_v8_3->src_prefix = pdu_v8_3->records[n].src_prefix;
            rec_v8_3->src_mask = pdu_v8_3->records[n].src_mask;
            rec_v8_3->src_as = pdu_v8_3->records[n].src_as;
            rec_v8_3->input = pdu_v8_3->records[n].input;

            if (bo == FT_HEADER_LITTLE_ENDIAN) {

              SWAPINT32(rec_v8_3->dFlows);
              SWAPINT32(rec_v8_3->dPkts);
              SWAPINT32(rec_v8_3->dOctets);
              SWAPINT32(rec_v8_3->First);
              SWAPINT32(rec_v8_3->Last);
              SWAPINT32(rec_v8_3->src_prefix);
              SWAPINT16(rec_v8_3->src_as);
              SWAPINT16(rec_v8_3->input);

            }

          } /* for */

          break;

        case 4:
          decode->rec_size = ftrec_size(decode->ver);
          pdu_v8_4 = (struct ftpdu_v8_4*) pdu;

          for (n = 0; n < pdu_v8_4->count; ++n) {

            rec_v8_4 = (struct fts2rec_v8_4*)
              (decode->buf + (n*decode->rec_size));

            rec_v8_4->unix_nsecs = ph->unix_nsecs;
            rec_v8_4->unix_secs = ph->unix_secs;
            rec_v8_4->sysUpTime = ph->sysUpTime;

            rec_v8_4->engine_type = pdu_v8_4->engine_type;
            rec_v8_4->engine_type = pdu_v8_4->engine_id;

            rec_v8_4->dFlows = pdu_v8_4->records[n].dFlows;
            rec_v8_4->dPkts = pdu_v8_4->records[n].dPkts;
            rec_v8_4->dOctets = pdu_v8_4->records[n].dOctets;
            rec_v8_4->First = pdu_v8_4->records[n].First;
            rec_v8_4->Last = pdu_v8_4->records[n].Last;
            rec_v8_4->dst_prefix = pdu_v8_4->records[n].dst_prefix;
            rec_v8_4->dst_mask = pdu_v8_4->records[n].dst_mask;
            rec_v8_4->dst_as = pdu_v8_4->records[n].dst_as;
            rec_v8_4->output = pdu_v8_4->records[n].output;

            if (bo == FT_HEADER_LITTLE_ENDIAN) {

              SWAPINT32(rec_v8_4->dFlows);
              SWAPINT32(rec_v8_4->dPkts);
              SWAPINT32(rec_v8_4->dOctets);
              SWAPINT32(rec_v8_4->First);
              SWAPINT32(rec_v8_4->Last);
              SWAPINT32(rec_v8_4->dst_prefix);
              SWAPINT16(rec_v8_4->dst_as);
              SWAPINT16(rec_v8_4->output);

            }

          } /* for */

          break;

        case 5:
          decode->rec_size = ftrec_size(decode->ver);
          pdu_v8_5 = (struct ftpdu_v8_5*) pdu;

          for (n = 0; n < pdu_v8_5->count; ++n) {

            rec_v8_5 = (struct fts2rec_v8_5*)
              (decode->buf + (n*decode->rec_size));

            rec_v8_5->unix_nsecs = ph->unix_nsecs;
            rec_v8_5->unix_secs = ph->unix_secs;
            rec_v8_5->sysUpTime = ph->sysUpTime;

            rec_v8_5->engine_type = pdu_v8_5->engine_type;
            rec_v8_5->engine_type = pdu_v8_5->engine_id;

            rec_v8_5->dFlows = pdu_v8_5->records[n].dFlows;
            rec_v8_5->dPkts = pdu_v8_5->records[n].dPkts;
            rec_v8_5->dOctets = pdu_v8_5->records[n].dOctets;
            rec_v8_5->First = pdu_v8_5->records[n].First;
            rec_v8_5->Last = pdu_v8_5->records[n].Last;
            rec_v8_5->src_prefix = pdu_v8_5->records[n].src_prefix;
            rec_v8_5->dst_prefix = pdu_v8_5->records[n].dst_prefix;
            rec_v8_5->src_mask = pdu_v8_5->records[n].src_mask;
            rec_v8_5->dst_mask = pdu_v8_5->records[n].dst_mask;
            rec_v8_5->src_as = pdu_v8_5->records[n].src_as;
            rec_v8_5->dst_as = pdu_v8_5->records[n].dst_as;
            rec_v8_5->input = pdu_v8_5->records[n].input;
            rec_v8_5->output = pdu_v8_5->records[n].output;

            if (bo == FT_HEADER_LITTLE_ENDIAN) {

              SWAPINT32(rec_v8_5->dFlows);
              SWAPINT32(rec_v8_5->dPkts);
              SWAPINT32(rec_v8_5->dOctets);
              SWAPINT32(rec_v8_5->First);
              SWAPINT32(rec_v8_5->Last);
              SWAPINT32(rec_v8_5->src_prefix);
              SWAPINT32(rec_v8_5->dst_prefix);
              SWAPINT16(rec_v8_5->src_as);
              SWAPINT16(rec_v8_5->dst_as);
              SWAPINT16(rec_v8_5->input);
              SWAPINT16(rec_v8_5->output);

            }

          } /* for */

          break;

        default:
          fprintf(stderr, "Unsupported agg_method\n");
          goto fts2rec_pdu_decode_out;
          break;

      } /* switch agg_method */
        
      break;

    default:
      break;

  } /* switch */

fts2rec_pdu_decode_out:

  decode->num = n;

  return n;

}

