/*
 * Copyright (c) 2001 Mark Fullmer and The Ohio State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      $Id: flow-send.c,v 1.5 2001/02/10 04:40:12 maf Exp $
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <fcntl.h>
#include "ftlib.h"
#include "support.h"

int debug;

void usage();

int main(argc, argv)
int argc;
char **argv;
{
  extern int errno;
  extern char *optarg;
  extern int optind;
  struct ftio ftio;
  struct ftprof ftp;
  struct ftver ftv;
  struct ftencode fte;
  int i, ret, tx_delay;
  char tmp_buf[256], *c;
  u_int32 ip_addr;
  int udp_port, udp_sock;
  struct sockaddr_in addr;
  void *rec;

  tx_delay = 1;
  bzero(&addr, sizeof (struct sockaddr_in));

  /* profile */
  ftprof_start (&ftp);

  while ((i = getopt(argc, argv, "d:h?x:")) != -1)
    switch (i) {

    case 'd': /* debug */
      debug = atoi(optarg);
      break;

    case 'h': /* help */
    case '?':
      usage();
      exit (0);
      break;

    case 'x': /* transmitter delay */
      tx_delay = atoi(optarg);
      break;

    default:
      usage();
      exit (1);
      break;

    } /* switch */

  /* expect IP address/port option */
  if ((argc - optind) != 1) {
    usage();
    exit (1);
  }

  if (strlen(argv[optind]) > 255) {
    fprintf(stderr, "ouch!\n");
    exit (1);
  }

  strcpy(tmp_buf, argv[optind]);

  /* skip to end or port seperator */
  for (c = tmp_buf; (*c != '/') && (*c); ++c);

  /* extract the port port */
  if (*c == '/') {
    *c = 0;
    ++c;
    udp_port = atoi(c);
  } else
    udp_port = FT_PORT;

  /* extract the ip address part */
  ip_addr = scan_ip(tmp_buf);

  /* create a socket for transmitting the flow pdu's */
  if ((udp_sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    fprintf(stderr, "socket(): %s\n", strerror(errno));
    return -1;
  }

  addr.sin_addr.s_addr = htonl(ip_addr);
  addr.sin_family = AF_INET;
  addr.sin_port = htons(udp_port);

  /* read from stdin */
  if (ftio_init(&ftio, 0, FT_IO_FLAG_READ) < 0) {
    fprintf(stderr, "ftio_init(): failed\n");
    exit (1);
  }

  /* initialize encode struct */
  ftencode_init(&fte);

  /* copy version from io stream */
  ftio_get_ver(&ftio, &ftv);
  bcopy(&ftv, &fte.ver, sizeof ftv);

  while ((rec = ftio_read(&ftio))) {

retry:

    ret = fts2rec_pdu_encode(&fte, rec);

/*   ret == 0 then send and clear out buffer
 *   ret > 0 then encode another
 *   ret < 0 then this encoding failed, send and clear out buffer
*/

    if (ret <= 0) {

      /* convert pdu to network byte order */
#if BYTE_ORDER == LITTLE_ENDIAN
      ftpdu_swap(&fte.buf, BYTE_ORDER);
#endif /* BYTE_ORDER == LITTLE_ENDIAN */

      if (sendto(udp_sock, &fte.buf, fte.buf_size, 0, (struct sockaddr*)&addr,
        sizeof (struct sockaddr_in)) == -1) {
        fprintf(stderr, "sendto(): %s\n", strerror(errno));
        exit (1);
      }

      /* try not to overload send/receive buffers */
      if (tx_delay)
        usleep((unsigned)tx_delay);

      /* reset encode buffer */
      ftencode_reset(&fte);

      /* if ret < 0 then the current record was not encoded */
      if (ret < 0)
        goto retry;

    }

  }

  /* any left over? */
  if (fte.buf_size) {

    /* convert pdu to network byte order */
    ftpdu_swap(&fte.buf, BYTE_ORDER);

    if (sendto(udp_sock, &fte.buf, fte.buf_size, 0, (struct sockaddr*)&addr,
      sizeof (struct sockaddr_in)) == -1) {
      fprintf(stderr, "sendto(): %s\n", strerror(errno));
      exit (1);
    }

    /* try not to overload send/receive buffers */
    if (tx_delay)
      usleep((unsigned)tx_delay);

  } /* fte.buf_size */
  
  if (debug > 0) {
    ftprof_end(&ftp, ftio_get_rec_total(&ftio));
    ftprof_print(&ftp, argv[0], stderr);
  }

  return ret;

} /* main */

void usage() {

  fprintf(stderr, "flow-send:\n\n");
  fprintf(stderr, " -d # set debug level.\n");
  fprintf(stderr, " -x # transmit delay in microseconds.\n");
  fprintf(stderr, " -V pdu version.\n");
  fprintf(stderr, " A.B.C.D/port\n");
  fprintf(stderr, "\n\n");

} /* usage */

